# Copyright (c) 2020 Mateus Dias Ribeiro
# Licensed under the MIT license
# Original file: 
# - https://github.com/mdribeiro/DeepCFD/blob/master/Models/UNetEx.py
#
# Copyright (c) 2023 Carlos Peña Monferrer
# Licensed under the MIT license
#
# Changes made:
# - Adding custom MaxUnpool for supporting ONNX export

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils import weight_norm
from .AutoEncoder import create_layer
from .MaxUnpool import *

def create_encoder_block(in_channels, out_channels, kernel_size, wn=True, bn=True,
                 activation=nn.ReLU, layers=2):
    encoder = []
    for i in range(layers):
        _in = out_channels
        _out = out_channels
        if i == 0:
            _in = in_channels
        encoder.append(create_layer(_in, _out, kernel_size, wn, bn, activation, nn.Conv2d))
    return nn.Sequential(*encoder)


def create_decoder_block(in_channels, out_channels, kernel_size, wn=True, bn=True,
                 activation=nn.ReLU, layers=2, final_layer=False):
    decoder = []
    for i in range(layers):
        _in = in_channels
        _out = in_channels
        _bn = bn
        _activation = activation
        if i == 0:
            _in = in_channels * 2
        if i == layers - 1:
            _out = out_channels
            if final_layer:
                _bn = False
                _activation = None
        decoder.append(create_layer(_in, _out, kernel_size, wn, _bn, _activation, nn.ConvTranspose2d))
    return nn.Sequential(*decoder)


def create_encoder(in_channels, filters, kernel_size, wn=True, bn=True, activation=nn.ReLU, layers=2):
    encoder = []
    for i in range(len(filters)):
        if i == 0:
            encoder_layer = create_encoder_block(in_channels, filters[i], kernel_size, wn, bn, activation, layers)
        else:
            encoder_layer = create_encoder_block(filters[i-1], filters[i], kernel_size, wn, bn, activation, layers)
        encoder = encoder + [encoder_layer]
    return nn.Sequential(*encoder)


def create_decoder(out_channels, filters, kernel_size, wn=True, bn=True, activation=nn.ReLU, layers=2):
    decoder = []
    for i in range(len(filters)):
        if i == 0:
            decoder_layer = create_decoder_block(filters[i], out_channels, kernel_size, wn, bn, activation, layers, final_layer=True)
        else:
            decoder_layer = create_decoder_block(filters[i], filters[i-1], kernel_size, wn, bn, activation, layers, final_layer=False)
        decoder = [decoder_layer] + decoder
    return nn.Sequential(*decoder)


class UNetEx(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, filters=[16, 32, 64], layers=3,
                 weight_norm=True, batch_norm=True, activation=nn.ReLU, final_activation=None):
        super().__init__()
        assert len(filters) > 0
        self.final_activation = final_activation
        self.encoder = create_encoder(in_channels, filters, kernel_size, weight_norm, batch_norm, activation, layers)
        decoders = []
        for i in range(out_channels):
            decoders.append(create_decoder(1, filters, kernel_size, weight_norm, batch_norm, activation, layers))
        self.decoders = nn.Sequential(*decoders)

    def encode(self, x):
        tensors = []
        indices = []
        sizes = []
        for encoder in self.encoder:
            x = encoder(x)
            sizes.append(x.size())
            tensors.append(x)
            x, ind = F.max_pool2d(x, 2, 2, return_indices=True)
            indices.append(ind)
        return x, tensors, indices, sizes

    def decode(self, _x, _tensors, _indices, _sizes):
        y = []
        max_unpool2d_onnx = MaxUnpool2d(kernel_size=2, stride=2, padding=0);
        max_unpool2d = nn.MaxUnpool2d(kernel_size=2, stride=2, padding=0);

        for _decoder in self.decoders:
            x = _x
            tensors = _tensors[:]
            indices = _indices[:]
            sizes = _sizes[:]
            for decoder in _decoder:
                tensor = tensors.pop()
                size = sizes.pop()
                ind = indices.pop()
                if torch.onnx.is_in_onnx_export():
                    x = max_unpool2d_onnx(x, ind, output_size=size)
                else:
                    x = max_unpool2d(x, ind, output_size=size)

                x = torch.cat([tensor, x], dim=1)
                x = decoder(x)
            y.append(x)
        return torch.cat(y, dim=1)

    def forward(self, x):
        x, tensors, indices, sizes = self.encode(x)
        x = self.decode(x, tensors, indices, sizes)
        if self.final_activation is not None:
            x = self.final_activation(x)
        return x



import os
import json
import torch
import pickle
import getopt
import sys
import onnx
import onnxruntime as ort
from onnx import ModelProto, StringStringEntryProto
from onnx import shape_inference
from onnx import checker
import torch.optim as optim
from torch.utils.data import TensorDataset
from torch.autograd import Variable


def parseOpts(argv):
    channels = ""
    nx = ""
    ny = ""
    kernel_size = ""
    filters = ""
    net = ""
    checkpoint = "checkpoint.pt"
    output = "checkpoint.onnx"
    metadata = {}

    try:
        opts, args = getopt.getopt(
                argv,"hc:,x:,y:k:f:n:i:o:m:",
            [
                "channels=",
                "nx=",
                "ny=",
                "kernel-size=",
                "filters=",
                "net=",
                "input=",
                "output=",
                "metadata="
            ]
        )
    except getopt.GetoptError as e:
        print("Invalid option. See python -m cfdonnx --help")
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h' or opt == '--help' or opt == '-help':
            print("cfdonnx "
                "\n    -c  --channels  number of channels (optional, read from state_dict['input_shape'] by default)"
                "\n    -x  --nx  X dimension (optional, read from state_dict['input_shape'] by default)"
                "\n    -y  --ny  Y dimension (optional, read from state_dict['input_shape'] by default )"
                "\n    -k  --kernel-size  kernel size (optional, read from state_dict['kernel_size] by default )"
                "\n    -f  --filters  filter size, e.g. 8,16,32,32 (optional, read from state_dict['filters'] by default )"
                "\n    -n  --net  Network architecture UNetEx or "
                    "AutoEncoder (default: UNetEx)"
                "\n    -i  --input Checkpoint (default: checkpoint.pt)"
                "\n    -o  --output  ONNX output file (default: checkpoint.onnx)"
                "\n    -m  --metadata  Metadata (default: none)"
                "\n    -h  --help  Display help"
                
            )
            sys.exit()
        elif opt in ("-c", "--channels"):
            channels = int(arg)
        elif opt in ("-x", "--nx"):
            nx = int(arg)
        elif opt in ("-y", "--ny"):
            ny = int(arg)
        elif opt in ("-k", "--kernel-size"):
            kernel_size = int(arg)
        elif opt in ("-f", "--filters"):
            filters = [int(x) for x in arg.split(',')]
        elif opt in ("-n", "--net"):
            print("net: ", arg)
            net = arg
        elif opt in ("-i", "--input"):
            checkpoint = arg
        elif opt in ("-o", "--output"):
            output = arg
        elif opt in ("-m", "--metadata"):
            metadata = arg

    options = {
        'channels': channels,
        'nx': nx,
        'ny': ny,
        'kernel_size': kernel_size,
        'filters': filters,
        'net': net,
        'checkpoint': checkpoint,
        'output': output,
        'metadata': metadata,
    }

    return options


def main():
    options = parseOpts(sys.argv[1:])
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    file_extension = os.path.splitext(options["checkpoint"])[1]

    if file_extension == '.pt':
        state_dict = torch.load(options["checkpoint"], map_location=device)

        net = None
        net_name = None
        kernel_size = None
        filters = None
        dummy_input = None

        if not options["net"]:
            net_name = state_dict["architecture"]
        else:
            net_name = options["net"]

        if (net_name == "UNetEx"):
            from .models.torch.UNetEx import UNetEx
            net = UNetEx
        else:
            print(net_name + " architecture not supported")
            
        if not options["kernel_size"]:
            kernel_size = state_dict["kernel_size"]
        else:
            kernel_size = options["kernel_size"]

        if not options["filters"]:
            filters = state_dict["filters"]
        else:
            filters = options["filters"]

        if not options["channels"] and not options["nx"] and not options["ny"]:
            input_shape = state_dict["input_shape"]
            dummy_input = torch.randn(input_shape).to(device)
        else:
            dummy_input = torch.randn(
                1,
                options["channels"],
                options["nx"],
                options["ny"]).to(device)

        model = net(
            3,
            3,
            filters=filters,
            kernel_size=kernel_size,
            batch_norm=False,
            weight_norm=False
        )
        model.load_state_dict(state_dict, strict=False)
        model.to(device)

        torch.onnx.export(
           model, 
           dummy_input,
           options["output"],
           input_names = ["inputModel"],
           export_params=True,
           output_names = ["outputModel"],
        )

    else:
       print(fileExtension + " files are not supported yet")

    ort_session = ort.InferenceSession(options["output"])

    model = onnx.load(options["output"])
    checker.check_model(model)

    metadata_dict = json.loads(options["metadata"])

    for key, value in metadata_dict.items():
        meta = StringStringEntryProto()
        meta.key = key
        meta.value = value
        meta = model.metadata_props.add().CopyFrom(meta)

    onnx.save(model, options["output"])


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='cfdonnx',
    version='0.1.0',
    description='Converting ML-CFD models to ONNX',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Carlos Pena Monferrer',
    keywords='cfd ml onnx',
    url='https://github.com/simzero/cfdonnx',
    packages=['cfdonnx', 'cfdonnx.models', 'cfdonnx.models.torch'],
    package_dir={'': 'src'},
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
    ],
)
